﻿using System;
using System.Linq;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Networks.IntegratedOptimization {
  public static class TtpUtils {
    public static double EvaluateTtp(TravelingSalesmanProblem tsp, int[] tour, BinaryKnapsackProblem ksp, bool[] loot, int[] availability, double rentingRatio, double minSpeed, double maxSpeed) {
      bool feasible;
      return EvaluateTtp(tsp, tour, ksp, loot, availability, rentingRatio, minSpeed, maxSpeed, out feasible);
    }

    public static double EvaluateTtp(TravelingSalesmanProblem tsp, int[] tour, BinaryKnapsackProblem ksp, bool[] loot, int[] availability, double rentingRatio, double minSpeed, double maxSpeed, out bool feasible) {
      double collectedWeight = 0.0;
      double objectiveValue = 0.0;
      double infeasibleBaseLine = -1000000.0;
      double speedCoefficient = (maxSpeed - minSpeed) / ksp.KnapsackCapacity.Value;

      int hideoutIdx = 0;
      while (tour[hideoutIdx] != 0) hideoutIdx++;
      int cityIdx = (hideoutIdx + 1) % tour.Length;
      int lastCityIdx = hideoutIdx;

      while (cityIdx != hideoutIdx) {
        double oldCollectedWeight = collectedWeight;
        var availableItems = availability.Select((c, i) => new { CityIdx = c, ItemIdx = i }).Where(x => x.CityIdx == tour[cityIdx]);

        foreach (var item in availableItems) {
          if (!loot[item.ItemIdx]) continue;
          collectedWeight += ksp.Weights[item.ItemIdx];
          objectiveValue += ksp.Values[item.ItemIdx];
        }

        objectiveValue -= Distance(tsp.Coordinates.CloneAsMatrix(), tour[lastCityIdx], tour[cityIdx]) * rentingRatio /
                          (maxSpeed - speedCoefficient * oldCollectedWeight);
        lastCityIdx = cityIdx;
        cityIdx = (cityIdx + 1) % tour.Length;
      }

      objectiveValue -= Distance(tsp.Coordinates.CloneAsMatrix(), tour[lastCityIdx], tour[hideoutIdx]) * rentingRatio /
                        (maxSpeed - speedCoefficient * collectedWeight);

      feasible = collectedWeight <= ksp.KnapsackCapacity.Value;
      if (!feasible) objectiveValue = infeasibleBaseLine - collectedWeight;

      return objectiveValue;
    }

    private static double Distance(double[,] coords, int fromIdx, int toIdx) {
      double fromX = coords[fromIdx, 0], fromY = coords[fromIdx, 1],
             toX = coords[toIdx, 0], toY = coords[toIdx, 1];
      return (int)Math.Ceiling(Math.Sqrt((toX - fromX) * (toX - fromX) + (toY - fromY) * (toY - fromY)));
    }
  }
}
